﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization.Operators.LCS;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.DecisionList {
  [StorableClass]
  [Item("DecisionList", "")]
  public class DecisionList : Item, IGAssistIndividual {
    [Storable]
    private IList<Rule> rules;
    public IList<Rule> Rules {
      get { return rules; }
    }

    [Storable]
    private IAction defaultAction;
    public IAction DefaultAction {
      get { return defaultAction; }
    }

    // default rule (action) is part of the rule set size
    public int RuleSetSize {
      get { return defaultAction == null ? rules.Count : rules.Count + 1; }
    }

    [StorableConstructor]
    protected DecisionList(bool deserializing) : base(deserializing) { }
    protected DecisionList(DecisionList original, Cloner cloner)
      : base(original, cloner) {
      defaultAction = original.DefaultAction;
      rules = new List<Rule>(original.rules.Count);
      foreach (var rule in original.Rules) {
        rules.Add(cloner.Clone(rule));
      }
    }
    public DecisionList()
      : base() {
      rules = new List<Rule>();
      defaultAction = null;
    }
    public DecisionList(List<Rule> rules)
      : base() {
      this.rules = rules;
      this.defaultAction = null;
    }
    public DecisionList(List<Rule> rules, IAction defaultAction)
      : base() {
      if (rules.Any(x => x.Action.Match(defaultAction))) { throw new ArgumentException("If a default action is used. The default action is not allowed be used in any rule."); }
      this.rules = rules;
      this.defaultAction = defaultAction;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new DecisionList(this, cloner);
    }

    // for convenience
    public IEnumerable<IAction> Evaluate(IEnumerable<DecisionListInput> input) {
      int numberOfAliveRules;
      return Evaluate(input, out numberOfAliveRules);
    }
    // for convenience
    public IEnumerable<IAction> Evaluate(IEnumerable<DecisionListInput> input, out int numberOfAliveRules) {
      double theoryLengtgh;
      return Evaluate(input, out numberOfAliveRules, out theoryLengtgh);
    }
    public IEnumerable<IAction> Evaluate(IEnumerable<DecisionListInput> input, out int numberOfAliveRules, out double theoryLengtgh) {
      var estimated = new List<IAction>();
      var activatedRules = new HashSet<Rule>();
      foreach (var dli in input) {
        foreach (var rule in rules) {
          if (rule.MatchInput(dli)) {
            estimated.Add(rule.Action);
            activatedRules.Add(rule);
            break;
          }
        }
      }
      numberOfAliveRules = activatedRules.Count;
      theoryLengtgh = activatedRules.Sum(x => x.ComputeTheoryLength());

      return estimated;
    }

    #region IGAssistIndividual Members

    public IGAssistNiche Niche {
      get { return defaultAction; }
    }

    public void SetNiche(IRandom random, IGAssistNiche niche) {
      var action = niche as IAction;
      if (action == null) {
        throw new ArgumentException("Niche has to be an action");
      }
      if (Niche != null) {
        throw new ArgumentException("Niche has already been set. It cannot be set again.");
      }
      defaultAction = action;
      var except = new List<IAction>() { action };
      foreach (var rule in rules.Where(x => x.Action.SameNiche(niche))) {
        rule.Action.Randomize(random, except);
      }
    }

    public void ApplySplit(IRandom random, double probability) {
      foreach (var rule in rules) {
        rule.ApplySplit(random, probability);
      }
    }

    public void ApplyMerge(IRandom random, double probability) {
      foreach (var rule in rules) {
        rule.ApplyMerge(random, probability);
      }
    }

    public void ApplyReinitialize(IRandom random, double probability, double oneProbability, IEnumerable<IDiscretizer> discretizers) {
      foreach (var rule in rules) {
        rule.ApplyReinitialize(random, probability, oneProbability, discretizers);
      }
    }
    #endregion
  }
}
